@@ -1,4 +1,4 @@
-#    aioimaplib : an IMAPrev4 lib using python asyncio 
+#    aioimaplib : an IMAPrev4 lib using python asyncio
 #    Copyright (C) 2016  Bruno Thomas
 #
 #    This program is free software: you can redistribute it and/or modify
@@ -23,9 +23,12 @@ import uuid
 from collections import deque
 from copy import deepcopy
 from datetime import datetime, timedelta
+from email._policybase import Compat32
 from email.header import Header
+from email.message import Message
 from functools import update_wrapper
 from math import ceil
+from operator import attrgetter
 
 from pytz import utc
 
@@ -38,11 +41,19 @@ sh.setFormatter(logging.Formatter("%(asctime)s %(levelname)s " +
 log.addHandler(sh)
 
 NONAUTH, AUTH, SELECTED, IDLE, LOGOUT = 'NONAUTH', 'AUTH', 'SELECTED', 'IDLE', 'LOGOUT'
-
 UID_RANGE_RE = re.compile(r'(?P<start>\d+):(?P<end>\d|\*)')
+CAPABILITIES = 'IDLE UIDPLUS MOVE ENABLE NAMESPACE'
+CRLF = b'\r\n'
+
+
+class InvalidUidSet(RuntimeError):
+    def __init__(self, *args) -> None:
+        super().__init__(*args)
 
 
 class ServerState(object):
+    DEFAULT_MAILBOXES = ['INBOX', 'Trash', 'Sent', 'Drafts']
+
     def __init__(self):
         self.mailboxes = dict()
         self.connections = dict()
@@ -60,27 +71,25 @@ class ServerState(object):
         if mailbox not in self.mailboxes[to]:
             self.mailboxes[to][mailbox] = list()
         m = deepcopy(mail)
-        m.id = self.max_id(to, mailbox) + 1
-        m.uid = self.max_uid(to) + 1
+        m.id = len(self.mailboxes[to][mailbox]) + 1
+        m.uid = self.max_uid(to, mailbox) + 1
         self.mailboxes[to][mailbox].append(m)
         return m.uid
 
-    def max_uid(self, user):
-        if user not in self.mailboxes: return 0
-        return max(map(lambda mailbox: max(mailbox, default=Mail(user), key=lambda msg: msg.uid).uid,
-                       self.mailboxes[user].values()))
+    def max_uid(self, user, mailbox):
+        if user not in self.mailboxes or mailbox not in self.mailboxes[user] \
+            or len(self.mailboxes[user][mailbox]) == 0: return 0
+        return max(self.mailboxes[user][mailbox], key=lambda msg: msg.uid).uid
 
     def max_id(self, user, mailbox):
-        if user not in self.mailboxes or \
-                mailbox not in self.mailboxes[user] or \
-                len(self.mailboxes[user][mailbox]) == 0:
-            return 0
-        return max(self.mailboxes[user][mailbox], key=lambda msg: msg.id).id
+        if user not in self.mailboxes or mailbox not in self.mailboxes[user]: return 0
+        return len(self.mailboxes[user][mailbox])
 
     def login(self, user_login, protocol):
         if user_login not in self.mailboxes:
             self.mailboxes[user_login] = dict()
-            self.mailboxes[user_login]['INBOX'] = list()
+        for mb in self.DEFAULT_MAILBOXES:
+            self.create_mailbox_if_not_exists(user_login, mb)
         if user_login not in self.connections:
             self.connections[user_login] = protocol
         if user_login not in self.subcriptions:
@@ -113,12 +122,15 @@ class ServerState(object):
         mb_re = re.compile(mailbox_search)
         return [mb for mb in self.subcriptions[user] if mb_re.match(mb)]
 
-    def list(self, user, mailbox_pattern):
+    def list(self, user, reference, mailbox_pattern):
+        mb = self.mailboxes[user]
+        for path_item in reference.split('/'):
+            mb = self.mailboxes[user].get(path_item, self.mailboxes[user])
         mb_re = re.compile(mailbox_pattern)
-        return sorted([mb for mb in self.mailboxes[user].keys() if mb_re.match(mb)])
+        return sorted([mb for mb in mb.keys() if mb_re.match(mb)])
 
     def remove(self, message, user, mailbox):
-        self.mailboxes[user][mailbox].remove(message)
+        self.remove_byid(user, mailbox, message.id)
 
     def delete_mailbox(self, user, mailbox):
         if mailbox in self.mailboxes[user]:
@@ -135,6 +147,25 @@ class ServerState(object):
             self.mailboxes[user][dest_mailbox] = list()
         self.mailboxes[user][dest_mailbox] += to_copy
 
+    def move(self, user, src_mailbox, dest_mailbox, id_range, msg_attribute):
+        id_getter = attrgetter(msg_attribute)
+        to_move = [msg for msg in self.mailboxes[user][src_mailbox] if id_getter(msg) in id_range]
+        id_moved = []
+        for msg in to_move:
+            self.remove(msg, user, src_mailbox)
+            id_moved.append(self.add_mail(user, msg, dest_mailbox))
+        if len(id_moved) == 0:
+            id_moved.append(0)
+        return range(min(id_moved), max(id_moved) + 1)
+
+    def remove_byid(self, user, mailbox, id):
+        msg = self.mailboxes[user][mailbox].pop(id-1)
+        self._reindex(user, mailbox)
+        return msg
+
+    def _reindex(self, user, mailbox):
+        for idx, msg in enumerate(self.mailboxes[user][mailbox]): msg.id = idx + 1
+
 
 def critical_section(next_state):
     @asyncio.coroutine
@@ -147,20 +178,24 @@ def critical_section(next_state):
 
     def decorator(func):
         def wrapper(self, *args, **kwargs):
-            asyncio.async(execute_section(self, next_state, func, *args, **kwargs))
+            asyncio.ensure_future(execute_section(self, next_state, func, *args, **kwargs))
 
         return update_wrapper(wrapper, func)
 
     return decorator
 
 
-command_re = re.compile(br'((DONE)|(?P<tag>\w+) (?P<cmd>[\w]+)([\w \.#@:\*"\(\)\{\}\[\]\+\-]+)?$)')
+command_re = re.compile(br'((DONE)|(?P<tag>\w+) (?P<cmd>[\w]+)([\w \.#@:\*"\(\)\{\}\[\]\+\-\\\%]+)?$)')
+FETCH_HEADERS_RE = re.compile(r'.*BODY.PEEK\[HEADER.FIELDS \((?P<headers>.+)\)\].*')
 
 
 class ImapProtocol(asyncio.Protocol):
     IDLE_STILL_HERE_PERIOD_SECONDS = 10
 
-    def __init__(self, server_state, fetch_chunk_size=0, loop=asyncio.get_event_loop()):
+    def __init__(self, server_state, fetch_chunk_size=0, capabilities=CAPABILITIES,
+                 loop=asyncio.get_event_loop()):
+        self.uidvalidity = int(datetime.now().timestamp())
+        self.capabilities = capabilities
         self.state_to_send = list()
         self.delay_seconds = 0
         self.loop = loop
@@ -217,7 +252,7 @@ class ImapProtocol(asyncio.Protocol):
 
     def send_raw_untagged_line(self, raw_response, continuation=False, max_chunk_size=0):
         prefix = b'+ ' if continuation else b'* '
-        raw_line = prefix + raw_response + b'\r\n'
+        raw_line = prefix + raw_response + CRLF
         if max_chunk_size:
             for nb_chunk in range(ceil(len(raw_line) / max_chunk_size)):
                 chunk_start_index = nb_chunk * max_chunk_size
@@ -236,7 +271,7 @@ class ImapProtocol(asyncio.Protocol):
     def login(self, tag, *args):
         self.user_login = args[0]
         self.server_state.login(self.user_login, self)
-        self.send_untagged_line('CAPABILITY IMAP4rev1')
+        self.send_untagged_line('CAPABILITY IMAP4rev1 %s' % self.capabilities)
         self.send_tagged_line(tag, 'OK LOGIN completed')
 
     @critical_section(next_state=LOGOUT)
@@ -288,8 +323,9 @@ class ImapProtocol(asyncio.Protocol):
         self.send_untagged_line('OK [PERMANENTFLAGS (\Answered \Flagged \Deleted \Seen \Draft \*)] Flags permitted.')
         self.send_untagged_line('{nb_messages} EXISTS'.format(nb_messages=len(mailbox)))
         self.send_untagged_line('{nb_messages} RECENT'.format(nb_messages=0))
-        self.send_untagged_line('OK [UIDVALIDITY 1400426466] UIDs valid')
-        self.send_untagged_line('OK [UIDNEXT {next_uid}] Predicted next UID'.format(next_uid=len(mailbox) + 1))
+        self.send_untagged_line('OK [UIDVALIDITY {uidvalidity}] UIDs valid'.format(uidvalidity=self.uidvalidity))
+        self.send_untagged_line('OK [UIDNEXT {next_uid}] Predicted next UID'.format(
+            next_uid=self.server_state.max_uid(self.user_login, mailbox_name) + 1))
         self.send_tagged_line(tag, 'OK [READ] Select completed (0.000 secs).')
 
     def search(self, tag, *args_param):
@@ -351,11 +387,11 @@ class ImapProtocol(asyncio.Protocol):
         arg_list = list(args)
         if arg_list[0] == 'uid':
             arg_list = list(args[1:])
-        uid = int(arg_list[0])  # args = ['12', '+FLAGS', 'FOO']
-        flag = arg_list[2]  # only support one flag and do not handle replacement (without + sign)
+        uid = int(arg_list[0])  # args = ['12', '+FLAGS', '(FOO)']
+        flags = ' '.join(arg_list[2:]).strip('()').split() # only support one flag and do not handle replacement (without + sign)
         for message in self.server_state.get_mailbox_messages(self.user_login, self.user_mailbox):
             if message.uid == uid:
-                message.flags.append(flag)
+                message.flags.extend(flags)
                 self.send_untagged_line('{uid} FETCH (UID {uid} FLAGS ({flags}))'.format(
                     uid=uid, flags=' '.join(message.flags)))
         self.send_tagged_line(tag, 'OK Store completed.')
@@ -366,7 +402,10 @@ class ImapProtocol(asyncio.Protocol):
         if arg_list[0] == 'uid':
             by_uid = True
             arg_list = list(args[1:])
-        fetch_range = self._build_fetch_range(arg_list[0])
+        try:
+            fetch_range = self._build_sequence_range(arg_list[0])
+        except InvalidUidSet:
+            return self.error(tag, 'Error in IMAP command: Invalid uidset')
         parts = arg_list[1:]
         parts_str = ' '.join(parts)
         for message in self.server_state.get_mailbox_messages(self.user_login, self.user_mailbox):
@@ -377,19 +416,27 @@ class ImapProtocol(asyncio.Protocol):
                 self.send_raw_untagged_line(response)
         self.send_tagged_line(tag, 'OK FETCH completed.')
 
-    def _build_fetch_range(self, uid_pattern):
-        range_re = re.compile(r'(\d):(\d|\*)')
+    def _build_sequence_range(self, uid_pattern):
+        range_re = re.compile(r'(\d+):(\d+|\*)')
         match = range_re.match(uid_pattern)
         if match:
+            start = int(match.group(1))
+            if start <= 0:
+                raise InvalidUidSet()
+
             if match.group(2) == '*':
-                return range(int(match.group(1)), sys.maxsize)
-            return range(int(match.group(1)), int(match.group(2)) + 1)
+                return range(start, sys.maxsize)
+
+            end = int(match.group(2))
+            if end <= 0 or end < start:
+                raise InvalidUidSet()
+            return range(start, end + 1)
         return [int(uid_pattern)]
 
     def _build_fetch_response(self, message, parts, by_uid=True):
         response = ('%d FETCH (UID %s' % (message.id, message.uid)).encode() if by_uid \
             else ('%d FETCH (' % message.id).encode()
-        for idx, part in enumerate(parts):
+        for part in parts:
             if part.startswith('(') or part.endswith(')'):
                 part = part.strip('()')
             if not response.endswith(b' ') and not response.endswith(b'('):
@@ -398,8 +445,18 @@ class ImapProtocol(asyncio.Protocol):
                 response += ('UID %s' % message.uid).encode()
             if part == 'BODY[]' or part == 'BODY.PEEK[]' or part == 'RFC822':
                 response += ('%s {%s}\r\n' % (part, len(message.as_bytes()))).encode() + message.as_bytes()
+            if part == 'BODY.PEEK[HEADER.FIELDS':
+                fetch_header = FETCH_HEADERS_RE.match(' '.join(parts))
+                if fetch_header:
+                    headers = fetch_header.group('headers')
+                    message_headers = Message(policy=Compat32(linesep='\r\n'))
+                    for hk in headers.split():
+                        message_headers[hk] = message.email.get(hk, '')
+                    response += ('BODY[HEADER.FIELDS (%s)] {%d}\r\n' %
+                                 (headers, len(message_headers.as_bytes()))).encode() + message_headers.as_bytes()
             if part == 'FLAGS':
                 response += ('FLAGS (%s)' % ' '.join(message.flags)).encode()
+        response = response.strip(b' ')
         response += b')'
         return response
 
@@ -411,34 +468,84 @@ class ImapProtocol(asyncio.Protocol):
 
     def append_literal(self, data):
         tag, mailbox_name, size = self.append_literal_command
-        if data == b'\r\n':
-            self.send_tagged_line(tag, 'OK APPEND completed.')
+        if data == CRLF:
+            if 'UIDPLUS' in self.capabilities:
+                self.send_tagged_line(tag, 'OK [APPENDUID %s %s] APPEND completed.' %
+                                      (self.uidvalidity, self.server_state.max_uid(self.user_login, mailbox_name)))
+            else:
+                self.send_tagged_line(tag, 'OK APPEND completed.')
             self.append_literal_command = None
             return
 
-        if len(data) != size:
+        literal_data, rest = data[:size], data[size:]
+        if len(literal_data) < size:
             self.send_tagged_line(self.append_literal_command[0],
-                                  'BAD literal length : expected %s but was %s' % (size, len(data)))
+                                  'BAD literal length : expected %s but was %s' % (size, len(literal_data)))
             self.append_literal_command = None
+        elif rest and rest != CRLF:
+            self.send_tagged_line(self.append_literal_command[0],
+                                  'BAD literal trailing data : expected CRLF but got %s' % (rest))
         else:
             m = email.message_from_bytes(data)
-            self.server_state.add_mail(m.get('To'), Mail(m), mailbox_name)
+            self.server_state.add_mail(self.user_login, Mail(m), mailbox_name)
+
+            if rest:
+                self.append_literal(rest)
 
     def expunge(self, tag, *args):
+        expunge_range = range(0, sys.maxsize)
+        uid_response = ''
+        if args and args[0] == 'uid':
+            uid_response = 'UID '
+            if len(args) > 1:
+                try:
+                    expunge_range = self._build_sequence_range(args[1])
+                except InvalidUidSet:
+                    return self.error(tag, 'Error in IMAP command: Invalid uidset')
         for message in self.server_state.get_mailbox_messages(self.user_login, self.user_mailbox).copy():
-            self.server_state.remove(message, self.user_login, self.user_mailbox)
-            self.send_untagged_line('{msg_uid} EXPUNGE'.format(msg_uid=message.uid))
-        self.send_tagged_line(tag, 'OK EXPUNGE completed.')
+            if message.uid in expunge_range:
+                self.server_state.remove(message, self.user_login, self.user_mailbox)
+                self.send_untagged_line('{msg_uid} EXPUNGE'.format(msg_uid=message.uid))
+        self.send_tagged_line(tag, 'OK %sEXPUNGE completed.' % uid_response)
 
     def capability(self, tag, *args):
-        self.send_untagged_line('CAPABILITY IMAP4rev1 LITERAL+ IDLE')
+        self.send_untagged_line('CAPABILITY IMAP4rev1 YESAUTH')
         self.send_tagged_line(tag, 'OK Pre-login capabilities listed, post-login capabilities have more')
 
+    def namespace(self, tag):
+        self.send_untagged_line('NAMESPACE (("" "/")) NIL NIL')
+        self.send_tagged_line(tag, 'OK NAMESPACE command completed')
+
+    def enable(self, tag, *args):
+        self.send_tagged_line(tag, 'OK %s enabled' % ' '.join(args))
+
     def copy(self, tag, *args):
         message_set, mailbox = args[0:-1], args[-1]
         self.server_state.copy(self.user_login, self.user_mailbox, mailbox, message_set)
         self.send_tagged_line(tag, 'OK COPY completed.')
 
+    def move(self, tag, *args):
+        args_list = list(args)
+        args_list.reverse()
+        msg_attribute = 'id'
+        if args[-1] == 'uid':
+            msg_attribute = 'uid'
+        mailbox, message_set = args_list[0:2]
+        seq_range = self._build_sequence_range(message_set)
+        seq_moved = self.server_state.move(self.user_login, self.user_mailbox, mailbox, seq_range, msg_attribute)
+        if 'UIDPLUS' in self.capabilities:
+            self.send_untagged_line(
+                'OK [COPYUID %d %d:%d %d:%d]' % (self.uidvalidity,
+                                                 seq_range.start, seq_range.stop-1,
+                                                 seq_moved.start, seq_moved.stop-1))
+        for msg_id in seq_moved:
+            self.send_untagged_line('{msg_id} EXPUNGE'.format(msg_id=msg_id))
+        self.send_tagged_line(tag, 'OK Done')
+
+    def id(self, tag, *args):
+        self.send_untagged_line('NIL')
+        self.send_tagged_line(tag, 'OK ID command completed')
+
     def noop(self, tag, *args):
         if len(self.state_to_send) > 0:
             for line in deque(self.state_to_send):
@@ -461,9 +568,9 @@ class ImapProtocol(asyncio.Protocol):
         if 'RECENT' in data_items:
             status_response += ' RECENT %s' % len([m for m in mailbox if 'RECENT' in m.flags])
         if 'UIDNEXT' in data_items:
-            status_response += ' UIDNEXT %s' % (self.server_state.max_uid(self.user_login) + 1)
+            status_response += ' UIDNEXT %s' % (self.server_state.max_uid(self.user_login, self.user_mailbox) + 1)
         if 'UIDVALIDITY' in data_items:
-            status_response += ' UIDVALIDITY %s' % (self.server_state.max_uid(self.user_login) + 1)
+            status_response += ' UIDVALIDITY %s' % self.uidvalidity
         if 'UNSEEN' in data_items:
             status_response += ' UNSEEN %s' % len([m for m in mailbox if 'UNSEEN' in m.flags])
         status_response += ')'
@@ -508,9 +615,10 @@ class ImapProtocol(asyncio.Protocol):
         self.send_tagged_line(tag, 'OK RENAME completed.')
 
     def list(self, tag, *args):
-        mailbox_pattern = args[0]
+        reference = args[0]
+        mailbox_pattern = args[1].replace('*', '.*').replace('%', '.*')
 
-        for mb in self.server_state.list(self.user_login, mailbox_pattern):
+        for mb in self.server_state.list(self.user_login, reference, mailbox_pattern):
             self.send_untagged_line('LIST () "/" %s' % mb)
         self.send_tagged_line(tag, 'OK LIST completed.')
 
@@ -531,9 +639,10 @@ class ImapProtocol(asyncio.Protocol):
 
 
 class MockImapServer(object):
-    def __init__(self, loop=None) -> None:
+    def __init__(self, capabilities=CAPABILITIES, loop=None) -> None:
         self._server_state = ServerState()
         self._connections = list()
+        self.capabilities = capabilities
         if loop is None:
             self.loop = asyncio.get_event_loop()
         else:
@@ -568,13 +677,14 @@ class MockImapServer(object):
     def get_connection(self, user):
         return self._server_state.get_connection(user)
 
-    def run_server(self, host='localhost', port=1143, fetch_chunk_size=0):
+    def run_server(self, host='127.0.0.1', port=1143, fetch_chunk_size=0, ssl_context=None):
         def create_protocol():
-            protocol = ImapProtocol(self._server_state, fetch_chunk_size, self.loop)
+            protocol = ImapProtocol(self._server_state, fetch_chunk_size, self.capabilities, self.loop)
             self._connections.append(protocol)
             return protocol
 
-        return self.loop.run_until_complete(self.loop.create_server(create_protocol, host, port))
+        server = self.loop.create_server(create_protocol, host, port, ssl=ssl_context)
+        return self.loop.run_until_complete(server)
 
     def reset(self):
         self._server_state.reset()
@@ -606,7 +716,8 @@ class Mail(object):
                message_id=None,
                quoted_printable=False,
                cc=None,
-               body_subtype='plain'
+               body_subtype='plain',
+               references=None
                ):
         """
         :param quoted_printable: boolean
@@ -620,6 +731,7 @@ class Mail(object):
         :param in_reply_to: str
         :param message_id: str
         :param body_subtype: str
+        :param references: list
         """
         charset = email.charset.Charset(encoding)
         msg = email.mime.nonmultipart.MIMENonMultipart('text', body_subtype, charset=encoding)
@@ -643,6 +755,9 @@ class Mail(object):
             msg['In-Reply-To'] = '<%s>' % in_reply_to
         if cc is not None:
             msg['Cc'] = ', '.join(cc)
+        if references is not None:
+            ' '.join(['<%s>' % ref for ref in references])
+
         return Mail(msg, date=date)
 
 if __name__ == '__main__':
